package gi;

import gi.SymbolSymbolMatrix.SymbolSymbolCell;

public final class ConstituentFilter {
	protected final int filterType;

	public ConstituentFilter(int filterType) {
		this.filterType = filterType;
	}

	/*
	 * We may want to assign values to filter types s.t. weak filters have small
	 * values and strong filters have large values.
	 */

	public static final int NULL_FILTER = 0;

	/**
	 * Tests whether a symbol pair is likely to be a constituent or not, using
	 * the type of constituent-filter specified by
	 * {@link ConstituentFilter#filterType}.
	 * 
	 * @param appearances
	 *            the set of pointers to all the appearances of the pair
	 *            (pointing to the 1st symbol).
	 * @return whether the pair is likely to be a constituent.
	 */
	public boolean isConstituent(SymbolSymbolCell appearances) {
		/*
		 * TODO Enhancement: The current implementation checks every input
		 * symbol pair based on its *current* appearances. A problem is that,
		 * the same symbol pair may be checked for more than once, with
		 * different sets of appearances, so it's possible that the results are
		 * different. A possible enhancement is to record every previous
		 * judgment and never check the same pair twice. Notice that in most
		 * cases, the first checking is based on more appearances than later
		 * checking, and is therefore more reliable.
		 */

		/*
		 * Caution: when the collection of appearances is too small to be
		 * statistically significant, for statistic filters we should just
		 * return a default result, e.g., true if we want to be conservative.
		 */

		switch (filterType) {
		case NULL_FILTER:
		default:
			return true;
		}
		// TODO implement a non-trivial constituent filter
	}

	/**
	 * @return whether the current filter type is a weak one (i.e., it doesn't
	 *         do much filtering).
	 */
	public boolean isWeak() {
		return filterType == 0; // TODO change to a proper value after
								// implementing more filter types
	}
}
