package gi;

import java.util.ArrayList;
import java.util.Arrays;

import gi.util.CYKParser;
import gi.util.SentenceGenerator;

public class GITesting {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		int times = 50; // repeating times
		int size = 200; // evaluation set size for EvalWeak()
		int gramComSize = 10; // set size for grammar combination
		double[] measures = { 0, 0, 0, 0 };
		double[][] m = new double[times][];
		for (int i = 0; i < times; i++) {
			// double[] m = CrossValidationWeak(args);
			m[i] = EvalWeak(args, size);
			// double[] m = CompareGrammarWeak(args, size);
			// double[] m = GrammarSelectingWeak(args, gramComSize, size,
			// gramComSize * times / 5);
			// double[] m = GrammarAveragingWeak(args, gramComSize, size,
			// gramComSize * times / 5);
			for (int j = 0; j < m[i].length; j++)
				measures[j] += m[i][j];
		}
		for (int j = 0; j < m[0].length; j++)
			measures[j] /= times;

		double[] stdev = new double[m[0].length];
		for (int i = 0; i < times; i++)
			for (int j = 0; j < stdev.length; j++)
				stdev[j] += (m[i][j] - measures[j]) * (m[i][j] - measures[j]);
		for (int j = 0; j < stdev.length; j++) {
			stdev[j] /= times - 1;
			stdev[j] = Math.sqrt(stdev[j]);
		}

		System.out.println("Precision: " + measures[0] + "\t+/- " + stdev[0]);
		System.out.println("Recall: " + measures[1] + "\t+/- " + stdev[1]);
		System.out.println("F-score: " + measures[2] + "\t+/- " + stdev[2]);
		if (measures[3] != 0)
			System.out.println("Posterior: " + measures[3] + "\t+/- "
					+ stdev[3]);
		System.out.println(Math.round(measures[0] * 1000) + "/"
				+ Math.round(measures[1] * 1000) + "/"
				+ Math.round(measures[2] * 1000));
	}

	/**
	 * Evaluate the grammar learning by cross-validation, comparing the weak
	 * generative capacity of the real grammar and the learned grammar, using
	 * criteria like precision, recall, F-score.
	 * 
	 * @param args
	 *            #0: the real grammar file; #1: the corpus generated from the
	 *            real grammar.
	 * @return the measures
	 */
	public static double[] CrossValidationWeak(String[] args) {
		Grammar gram = new Grammar(args[0]);
		Corpus corp = new Corpus(args[1]);

		CYKParser parser = new CYKParser();
		SentenceGenerator generator = new SentenceGenerator();

		double precision = 0, recall = 0, fscore = 0;
		double posterior = 0;

		int n = 10; // n-fold cross-validation

		int[] boundary = new int[n + 1];
		boundary[0] = 0;
		int size = corp.sentences.length / n;
		for (int i = 1; i < n; i++) {
			boundary[i] = size * i;
		}
		boundary[n] = corp.sentences.length;
		for (int i = 0; i < n; i++) {
			Corpus training = corp.copyExceptPart(boundary[i], boundary[i + 1]);

			// Learning
			GrammarLearning gl = new GrammarLearning();
			Grammar g = gl.grammarLearning(training);

			// Measure Recall
			parser.setGrammar(g);
			int count = 0;
			for (int j = boundary[i]; j < boundary[i + 1]; j++) {
				if (parser.parse(corp.sentences[j]))
					count++;
			}
			double r = ((double) count) / (boundary[i + 1] - boundary[i]);
			recall += r;

			// Measure Precision
			generator.setGrammar(g);
			parser.setGrammar(gram);
			count = 0;
			for (int j = 0; j < size; j++) {
				Sentence tmp = generator.generate();
				if (parser.parse(tmp))
					count++;
			}
			double p = ((double) count) / size;
			precision += p;

			// F-Score
			double f = 2 * p * r / (p + r);
			fscore += f;

			// Posterior
			posterior += gl.totalPosteriorGain;

			// if (r < 0.75 || p < 0.6)
			// System.out.println("R: " + r + "\tP: " + p);
			System.out.println(p + "\t" + r + "\t" + f + "\t"
					+ gl.totalPosteriorGain);
		}

		precision /= n;
		recall /= n;
		fscore /= n;
		posterior /= n;
		// System.out.println(n + "-fold cross-validation:");
		// System.out.println("Precision: " + precision);
		// System.out.println("Recall: " + recall);
		// System.out.println("F-score: " + fscore);

		double[] re = { precision, recall, fscore, posterior };
		return re;
	}

	/**
	 * Evaluate the grammar learning, comparing the weak generative capacity of
	 * the real grammar and the learned grammar, using criteria like precision,
	 * recall, F-score.
	 * 
	 * @param args
	 *            #0: the real grammar file; #1: the corpus generated from the
	 *            real grammar.
	 * @param size
	 *            the size of testing sets
	 * @return the measures
	 */
	public static double[] EvalWeak(String[] args, int size) {
		Grammar gram = new Grammar(args[0]);
		Corpus corp = new Corpus(args[1]);

		CYKParser parser = new CYKParser();
		SentenceGenerator generator = new SentenceGenerator();

		// Learning
		GrammarLearning gl = new GrammarLearning();
		Grammar g = gl.grammarLearning(corp);

		// Measure Recall
		generator.setGrammar(gram);
		parser.setGrammar(g);
		int count = 0;
		for (int j = 0; j < size; j++) {
			Sentence tmp = generator.generate();
			if (parser.parse(tmp))
				count++;
		}
		double recall = ((double) count) / size;

		// Measure Precision
		generator.setGrammar(g);
		parser.setGrammar(gram);
		count = 0;
		for (int j = 0; j < size; j++) {
			Sentence tmp = generator.generate();
			if (parser.parse(tmp))
				count++;
		}
		double precision = ((double) count) / size;

		// F-Score
		double fscore = 2 * precision * recall / (precision + recall);

		// if (recall < 0.99 || precision < 0.99)
		// System.out.println("R: " + recall + "\tP: " + precision);
		System.out.println(precision + "\t" + recall + "\t" + fscore + "\t"
				+ gl.totalPosteriorGain);

		double[] re = { precision, recall, fscore, gl.totalPosteriorGain };
		return re;
	}

	/**
	 * @param args
	 *            #0: the real grammar file; #1: the learned grammar file
	 * @param size
	 *            the size of testing sets
	 * @return the measures
	 */
	public static double[] CompareGrammarWeak(String[] args, int size) {
		Grammar tgram = new Grammar(args[0]);
		Grammar lgram = new Grammar(args[1]);

		CYKParser parser = new CYKParser();
		SentenceGenerator generator = new SentenceGenerator();

		// Measure Recall
		generator.setGrammar(tgram);
		parser.setGrammar(lgram);
		int count = 0;
		for (int j = 0; j < size; j++) {
			Sentence tmp = generator.generate();
			if (parser.parse(tmp))
				count++;
		}
		double recall = ((double) count) / size;

		// Measure Precision
		generator.setGrammar(lgram);
		parser.setGrammar(tgram);
		count = 0;
		for (int j = 0; j < size; j++) {
			Sentence tmp = generator.generate();
			if (parser.parse(tmp))
				count++;
		}
		double precision = ((double) count) / size;

		// F-Score
		double fscore = 2 * precision * recall / (precision + recall);

		// if (recall < 0.75 || precision < 0.6)
		System.out.println(precision + "\t" + recall + "\t" + fscore);

		double[] re = { precision, recall, fscore };
		return re;
	}

	/* used for grammar selection/average */
	protected static ArrayList<Grammar> gramList;
	protected static ArrayList<double[]> measureList;

	/**
	 * Grammar selection by posterior. For speed consideration, we create a pool
	 * of grammars and in each test randomly fetch a set from it for grammar
	 * selection, instead of actually creating new grammars in each test.
	 * 
	 * @param args
	 *            #0: the real grammar file; #1: the training corpus
	 * @param num
	 *            the number of grammars to be learned in each test, among which
	 *            the one with the highest posterior will be selected
	 * @param size
	 *            the size of testing sets
	 * @param poolsize
	 *            the size of the grammar pool
	 * @return the measures
	 */
	public static double[] GrammarSelectingWeak(String[] args, int num,
			int size, int poolsize) {
		CYKParser parser = new CYKParser();
		SentenceGenerator generator = new SentenceGenerator();
		Grammar gram = new Grammar(args[0]);

		if (gramList == null) { // first time calling
			// create a list of grammars
			gramList = new ArrayList<Grammar>(poolsize);
			measureList = new ArrayList<double[]>(poolsize);
			for (int i = 0; i < poolsize; i++) {
				Corpus corp = new Corpus(args[1]);

				// Learning
				GrammarLearning gl = new GrammarLearning();
				Grammar g = gl.grammarLearning(corp);

				// Measure Recall
				generator.setGrammar(gram);
				parser.setGrammar(g);
				int count = 0;
				for (int j = 0; j < size; j++) {
					Sentence tmp = generator.generate();
					if (parser.parse(tmp))
						count++;
				}
				double recall = ((double) count) / size;

				// Measure Precision
				generator.setGrammar(g);
				parser.setGrammar(gram);
				count = 0;
				for (int j = 0; j < size; j++) {
					Sentence tmp = generator.generate();
					if (parser.parse(tmp))
						count++;
				}
				double precision = ((double) count) / size;

				// F-Score
				double fscore = 2 * precision * recall / (precision + recall);

				double[] re = { precision, recall, fscore,
						gl.totalPosteriorGain };

				gramList.add(g);
				measureList.add(re);
			}
		}

		// do selection
		double posterior = Double.NEGATIVE_INFINITY;
		double[] re = null;
		for (int i = 0; i < num; i++) {
			int ind = (int) (Math.random() * poolsize);
			if (measureList.get(ind)[3] > posterior) {
				re = measureList.get(ind);
				posterior = re[3];
			}
		}

		System.out.println(re[0] + "\t" + re[1] + "\t" + re[2] + "\t" + re[3]);

		return re;
	}

	/**
	 * Grammar average by posterior. For speed consideration, we create a pool
	 * of grammars and in each test randomly fetch a set from it for grammar
	 * average, instead of actually creating new grammars in each test.
	 * 
	 * For generating sentences, we randomly select a grammar by its weight,
	 * generate a sentence, judge by averaging parsing, and if yes, use it.
	 * 
	 * This method actually defines such a p-language: First, make a mixture of
	 * languages with the (normalized) posterior; then the predicting by average
	 * with threshold actually defines a subset of the language union, with the
	 * probabilities inside unchanged; to make it a correct p-language,
	 * normalize it.
	 * 
	 * @param args
	 *            #0: the real grammar file; #1: the training corpus
	 * @param num
	 *            the number of grammars to be learned, among which the one with
	 *            the highest posterior will be selected
	 * @param size
	 *            the size of testing sets
	 * @param poolsize
	 *            the size of the grammar pool
	 * @return the measures
	 */
	public static double[] GrammarAveragingWeak(String[] args, int num,
			int size, int poolsize) {
		Grammar gram = new Grammar(args[0]);
		CYKParser parser = new CYKParser();
		SentenceGenerator generator = new SentenceGenerator();

		if (gramList == null) { // first time calling
			// create a list of grammars
			gramList = new ArrayList<Grammar>(poolsize);
			measureList = new ArrayList<double[]>(poolsize);
			for (int i = 0; i < poolsize; i++) {
				Corpus corp = new Corpus(args[1]);

				// Learning
				GrammarLearning gl = new GrammarLearning();
				Grammar g = gl.grammarLearning(corp);

				double[] re = { 0, 0, 0, gl.totalPosteriorGain };
				gramList.add(g);
				measureList.add(re);
			}
		}

		// Selection
		Grammar[] gramSet = new Grammar[num];
		double[] postgains = new double[num];
		double[] weights = new double[num];
		for (int i = 0; i < num; i++) {
			int ind = (int) (Math.random() * poolsize);
			gramSet[i] = gramList.get(ind);
			postgains[i] = measureList.get(ind)[3];
		}
		double sum = 0;
		for (int i = 0; i < num; i++) {
			sum += postgains[i];
		}
		double mean = sum / num;
		double stdev = 0;
		for (int i = 0; i < num; i++) {
			double dif = postgains[i] - mean;
			stdev += dif * dif;
		}
		stdev /= num;
		stdev = Math.sqrt(stdev);
		sum = 0;
		for (int i = 0; i < num; i++) {
			weights[i] = Math.exp((postgains[i] - mean) / stdev);
			sum += weights[i];
		}
		for (int i = 0; i < num; i++) {
			weights[i] /= sum;
			if (weights[i] > 0.5)
				System.out.println("Found a dominant grammar.");
		}

		// Measure Recall
		generator.setGrammar(gram);
		int count = 0;
		for (int j = 0; j < size; j++) {
			Sentence tmp = generator.generate();
			if (GrammarAveragingParse(tmp, gramSet, weights))
				count++;
		}
		double recall = ((double) count) / size;

		// Measure Precision
		double[] weightSums = new double[num];
		sum = 0;
		for (int i = 0; i < num; i++) {
			sum += weights[i];
			weightSums[i] = sum;
		}
		parser.setGrammar(gram);
		count = 0;
		for (int j = 0; j < size; j++) {
			// generte a sentence
			Sentence tmp;
			do {
				double rand = Math.random() * sum;
				int ind = Arrays.binarySearch(weightSums, 0, num, rand);
				if (ind < 0)
					ind = -ind - 1;
				generator.setGrammar(gramSet[ind]);
				tmp = generator.generate();
			} while (!GrammarAveragingParse(tmp, gramSet, weights));

			if (parser.parse(tmp))
				count++;
		}
		double precision = ((double) count) / size;

		// F-Score
		double fscore = 2 * precision * recall / (precision + recall);

		// Average Posterior Gain (no statistic meaning)
		double avgPostGain = 0;
		for (int i = 0; i < num; i++) {
			avgPostGain += postgains[i] * weights[i];
		}

		double[] re = { precision, recall, fscore, avgPostGain };

		System.out.println(re[0] + "\t" + re[1] + "\t" + re[2] + "\t" + re[3]);

		return re;
	}

	/**
	 * Parse by grammar averaging. Notice that there are two parameters in the
	 * implementation. A threshold (0.01) is used to discard grammars that have
	 * too small weights to have any influence. Another threshold (0.5) is used
	 * tp pass or deny the vote.
	 * 
	 * @param s
	 * @param gramSet
	 * @param weights
	 * @return
	 */
	protected static boolean GrammarAveragingParse(Sentence s,
			Grammar[] gramSet, double[] weights) {
		CYKParser parser = new CYKParser();
		double vote = 0;
		for (int i = 0; i < weights.length; i++) {
			if (weights[i] < 0.01 / weights.length)
				continue; // too small to have any influence

			parser.setGrammar(gramSet[i]);
			if (parser.parse(s))
				vote += weights[i];
		}
		if (vote > 0.5)
			return true;
		else
			return false;
	}
}
