package gi;

import gi.util.CNFConverter;
import gi.util.CYKParser;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.regex.MatchResult;

public class Tmp {
	public static void main(String[] args) {
		// Corpus c = new Corpus("tmp/wsj2k.cp");
		// c.saveToReadableFile("Data/TA1_2000a.cpr");
		// Grammar g = new Grammar("tmp/tmp.g");
		// g.saveToReadableFile2("Data/emile");

		// testSelAvg();

		// adiosResult();

		// testEmileGrammar("TA1");

		String[] args2 = { "tmp/atis.cp", "tmp/tmp.parse", "0.5", "0.5", "1.4" };
		learnToBracketing(args2);
		try {
			String s = null;
			Process p = Runtime.getRuntime().exec(
					"tmp/evalb -p tmp/evalb.prm tmp/atis.bkt tmp/tmp.parse");
			BufferedReader stdInput = new BufferedReader(new InputStreamReader(
					p.getInputStream()));
			while ((s = stdInput.readLine()) != null) {
				System.out.println(s);
			}

		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * @param args
	 *            0: corpus file name; 1: output bracketing file name; 2: delta;
	 *            3: delta2; 4: alpha.
	 */
	public static void learnToBracketing(String[] args) {
		GrammarLearning.delta = Double.parseDouble(args[2]);
		GrammarLearning.delta2 = Double.parseDouble(args[3]);
		GrammarLearning.alpha = Double.parseDouble(args[4]);

		Corpus c = new Corpus(args[0]);
		Corpus c2 = c.copy();
		GrammarLearning gl = new GrammarLearning();
		Grammar g = gl.grammarLearning(c);

		CYKParser parser = new CYKParser();
		parser.setGrammar(g);
		int count = 0;
		try {
			FileWriter fw = new FileWriter(args[1]);
			for (int i = 0; i < c.sentences.length; i++) {
				Sentence s = c2.sentences[i];
				if (parser.parse(s)) {
					count++;
					fw.write(parser.getBestBracketingString() + "\n");
				} else {
					s = c.sentences[i];
					fw.write(parser.getPartialBracketingString(s) + "\n");
				}
			}
			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		System.out.println("Recall: " + count + "/" + c.sentences.length);
		System.out.println("Posterior: " + gl.totalPosteriorGain);
	}

	public static void printStat(double[][] data) {
		double[] measures = new double[data[0].length];
		for (int i = 0; i < data.length; i++)
			for (int j = 0; j < data[i].length; j++)
				measures[j] += data[i][j];
		for (int j = 0; j < measures.length; j++)
			measures[j] /= data.length;

		double[] stdev = new double[data[0].length];
		for (int i = 0; i < data.length; i++)
			for (int j = 0; j < stdev.length; j++)
				stdev[j] += (data[i][j] - measures[j])
						* (data[i][j] - measures[j]);
		for (int j = 0; j < stdev.length; j++) {
			stdev[j] /= data.length - 1;
			stdev[j] = Math.sqrt(stdev[j]);
		}

		System.out.println("Precision: " + measures[0] + "\t+/- " + stdev[0]);
		System.out.println("Recall: " + measures[1] + "\t+/- " + stdev[1]);
		System.out.println("F-score: " + measures[2] + "\t+/- " + stdev[2]);
		if (measures.length > 3)
			System.out.println("Posterior: " + measures[3] + "\t+/- "
					+ stdev[3]);
		System.out.println(Math.round(measures[0] * 1000) + "/"
				+ Math.round(measures[1] * 1000) + "/"
				+ Math.round(measures[2] * 1000));
	}

	public static void testSelAvg() {
		String[] args;
		// 10-sel
		args = new String[2];
		args[0] = "Data/test2.g";
		args[1] = "Data/test2.cp";
		int times = 50; // repeating times
		int size = 200; // evaluation set size for EvalWeak()
		int gramComSize = 10; // set size for grammar combination
		double[][] m = new double[times][];
		for (int i = 0; i < times; i++) {
			m[i] = GITesting.GrammarSelectingWeak(args, gramComSize, size,
					gramComSize * times / 5);
		}
		printStat(m);

		System.out.println("");

		// 10-avg
		for (int i = 0; i < times; i++) {
			m[i] = GITesting.GrammarAveragingWeak(args, gramComSize, size,
					gramComSize * times / 5);
		}
		printStat(m);

		System.out.println("");

		// 100 tests avg
		m = GITesting.measureList.toArray(new double[0][]);
		printStat(m);
	}

	public static void adiosResult() {
		int times = 30; // repeating times
		Scanner s = new Scanner(System.in);
		double measure[][] = new double[times][3];
		for (int i = 0; i < times; i++) {
			Scanner ss = new Scanner(s.nextLine());
			ss
					.findInLine("Recall =  Accepted: \\d+ out of \\d+ \\((.+)\\) Precision =  Accepted: \\d+ out of \\d+ \\((.+)\\)");
			MatchResult m = ss.match();
			double r = measure[i][1] = Double.valueOf(m.group(1));
			double p = measure[i][0] = Double.valueOf(m.group(2));
			measure[i][2] = 2 * p * r / (p + r);
			System.out.println("p:\t" + p + "\tr:\t" + r);
		}
		printStat(measure);
	}

	public static void readEmileGrammar(String fn, String fn2) {
		try {
			Scanner s = new Scanner(new BufferedReader(new FileReader(fn)));
			s.nextLine();
			s.nextLine();
			s.nextLine();

			ArrayList<ArrayList<String>> gramBuf = new ArrayList<ArrayList<String>>();
			while (s.hasNext()) {
				Scanner ss = new Scanner(s.nextLine());
				ArrayList<String> buf = new ArrayList<String>();
				buf.add(ss.findInLine("\\d+"));
				ss.next();
				while (ss.hasNext()) {
					String str = ss.findInLine("\\[\\d+\\]|\\w+|\\.");
					if (str.charAt(0) == '.')
						continue;
					if (str.charAt(0) == '[')
						str = str.substring(1, str.length() - 1);
					else
						str = '+' + str;
					buf.add(str);
				}
				if (buf.size() > 1)
					gramBuf.add(buf);
			}
			s.close();

			FileWriter fw = new FileWriter("tmp/$tmp");
			fw.write("(\n");
			String cur = null;
			for (ArrayList<String> arr : gramBuf) {
				if (!arr.get(0).equals(cur)) {
					if (cur != null)
						fw.write(")\n");
					fw.write("(" + arr.get(0) + "\t->");
					cur = arr.get(0);
				}
				fw.write("\t(1");
				for (int i = 1; i < arr.size(); i++) {
					fw.write(" " + arr.get(i));
				}
				fw.write(")\n");
			}
			fw.write(")");
			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		String[] args = { "tmp/$tmp", fn2 };
		CNFConverter.main(args);
	}

	public static void testEmileGrammar(String fn) {
		int times = 50; // repeating times
		int size = 200; // evaluation set size for EvalWeak()
		double[][] m = new double[times][];
		for (int i = 0; i < times; i++) {
			readEmileGrammar("tmp/" + fn + "_2000a.g." + (i + 1), "tmp/tmp.g");
			String[] args = { "Data/" + fn + ".g", "tmp/tmp.g" };
			m[i] = GITesting.CompareGrammarWeak(args, size);
		}
		printStat(m);
	}
}
