package gi.util;

import gi.AndNode;
import gi.Grammar;
import gi.OrNode;
import gi.Symbol;
import gi.Term;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Scanner;
import java.util.regex.MatchResult;

/**
 * Convert a CFG to CNF. The input CFG file must be formatted.
 * 
 * @author Kewei Tu
 * 
 */
public class CNFConverter {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		CNFConverter c = new CNFConverter();
		c.ReadGrammar(args[0]);
		if (c.g != null)
			c.g.saveToFile(args[1]);
	}

	protected HashMap<String, OrNode> orMap = new HashMap<String, OrNode>();
	protected ArrayList<OrNode> orList = new ArrayList<OrNode>();
	protected HashMap<String, HashMap<String, AndNode>> andMap = new HashMap<String, HashMap<String, AndNode>>();
	protected ArrayList<AndNode> andList = new ArrayList<AndNode>();
	protected HashMap<String, Term> termMap = new HashMap<String, Term>();
	protected Grammar g;

	protected void ReadGrammar(String filename) {
		ArrayList<Object[]> nt2nt = new ArrayList<Object[]>();
		HashMap<OrNode, int[]> nt2ntNum = new HashMap<OrNode, int[]>();

		try {
			Scanner s = new Scanner(
					new BufferedReader(new FileReader(filename)));
			s.skip("\\s*\\(");

			OrNode startSymbol = null;
			while (s.hasNext()) {
				s.skip("\\s*([\\(\\)])");
				String paren = s.match().group(1);
				if (paren.equals(")"))
					break;
				String nt = s.next();
				OrNode or = orMap.get(nt);
				if (or == null) {
					or = new OrNode(new ArrayList<Symbol>(),
							new ArrayList<double[]>());
					orMap.put(nt, or);
					orList.add(or);
				}

				if (startSymbol == null)
					startSymbol = or;

				s.skip("\\s*->");

				while (true) {
					s.skip("\\s*([\\(\\)])");
					paren = s.match().group(1);
					if (paren.equals(")"))
						break;
					s.findWithinHorizon("(\\d+)\\s*([\\w|\\s|\\+]*)\\)", 0);
					MatchResult m = s.match();
					int weight = Integer.valueOf(m.group(1));
					String rhs = m.group(2);

					Symbol sym;
					Scanner rhss = new Scanner(rhs);
					ArrayList<String> list = new ArrayList<String>();
					while (rhss.hasNext())
						list.add(rhss.next());

					if (list.size() == 1) {
						String str = list.get(0);
						if (str.startsWith("+")) {
							// terminal
							str = str.substring(1);
							Term t = termMap.get(str);
							if (t == null) {
								t = new Term(str);
								termMap.put(str, t);
							}
							sym = t;
						} else {
							// non-terminal
							Object[] tmp = { or, str, weight };
							nt2nt.add(tmp);
							int[] i = nt2ntNum.get(or);
							if (i == null) {
								int[] n = { 1 };
								nt2ntNum.put(or, n);
							} else
								i[0]++;
							continue;
						}
					} else
						sym = constructCNFTree(list, 0, list.size());

					or.children.add(sym);
					double[] tmp = { weight };
					or.weights.add(tmp);
					sym.parents.add(or);
				}
			}

			s.close();

			// process nt2nt
			while (!nt2nt.isEmpty()) {
				Object[] a = nt2nt.remove(0);
				OrNode or = (OrNode) a[0];
				OrNode or2 = orMap.get(a[1]);
				int coef = (Integer) a[2];
				if (or2 == null)
					throw new Exception("Fatal error in reading the grammar: "
							+ a[1] + " not defined!");
				if (nt2ntNum.containsKey(or2)) {
					nt2nt.add(a);
					continue;
				}

				double sum = 0;
				for (double[] d : or2.weights) {
					sum += d[0];
				}
				for (int i = 0; i < or2.children.size(); i++) {
					or.children.add(or2.children.get(i));
					double[] tmp = { or2.weights.get(i)[0] * coef / sum };
					or.weights.add(tmp);
				}

				int[] n = nt2ntNum.get(or);
				if (n[0] == 1)
					nt2ntNum.remove(or);
				else
					n[0]--;
			}

			g = new Grammar(termMap.values().toArray(new Term[0]));
			for (AndNode and : andList) {
				g.andNodes.add(and);
			}
			for (OrNode or : orList) {
				g.orNodes.add(or);
			}
			g.startSymbol = startSymbol;
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * @param list
	 * @param start
	 *            inclusive start position
	 * @param end
	 *            exclusive end position
	 * @return
	 */
	protected AndNode constructCNFTree(ArrayList<String> list, int start,
			int end) {
		OrNode left, right;
		AndNode and = null;
		if (end - start > 3) {
			int seg = (end - start) / 2;

			AndNode land = constructCNFTree(list, start, start + seg);
			ArrayList<Symbol> chd = new ArrayList<Symbol>();
			chd.add(land);
			ArrayList<double[]> wt = new ArrayList<double[]>();
			double[] tmp = { 1 };
			wt.add(tmp);
			left = new OrNode(chd, wt);
			orList.add(left);

			AndNode rand = constructCNFTree(list, start + seg, end);
			chd = new ArrayList<Symbol>();
			chd.add(rand);
			wt = new ArrayList<double[]>();
			double[] tmp2 = { 1 };
			wt.add(tmp2);
			right = new OrNode(chd, wt);
			orList.add(right);

			and = new AndNode(left, right, 0);
		} else if (end - start == 3) {
			left = String2OrNode(list.get(start));

			AndNode rand = constructCNFTree(list, start + 1, end);
			ArrayList<Symbol> chd = new ArrayList<Symbol>();
			chd.add(rand);
			ArrayList<double[]> wt = new ArrayList<double[]>();
			double[] tmp2 = { 1 };
			wt.add(tmp2);
			right = new OrNode(chd, wt);
			orList.add(right);

			and = new AndNode(left, right, 0);
		} else { // end-start==2
			HashMap<String, AndNode> tmp = andMap.get(list.get(start));
			if (tmp != null)
				and = tmp.get(list.get(start + 1));
			if (and == null) {
				left = String2OrNode(list.get(start));
				right = String2OrNode(list.get(start + 1));
				and = new AndNode(left, right, 0);

				if (tmp == null) {
					tmp = new HashMap<String, AndNode>();
					andMap.put(list.get(start), tmp);
				}
				tmp.put(list.get(start + 1), and);
			}
		}

		andList.add(and);
		return and;
	}

	protected OrNode String2OrNode(String str) {
		OrNode or = orMap.get(str);
		if (or == null) {
			if (str.startsWith("+")) {
				// terminal
				String str2 = str.substring(1);
				Term t = termMap.get(str2);
				if (t == null) {
					t = new Term(str2);
					termMap.put(str2, t);
				}
				ArrayList<Symbol> chd = new ArrayList<Symbol>();
				chd.add(t);
				ArrayList<double[]> wt = new ArrayList<double[]>();
				double[] tmp = { 1 };
				wt.add(tmp);
				or = new OrNode(chd, wt);
			} else {
				// OrNode
				or = new OrNode(new ArrayList<Symbol>(),
						new ArrayList<double[]>());
			}
			orMap.put(str, or);
			orList.add(or);
		}
		return or;
	}
}
