package gi.util;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Scanner;

/**
 * Convert corpus files to my format.
 * 
 * @author Kewei Tu
 * 
 */
public class CorpusFormatter {

	/**
	 * @param args
	 *            The first parameter "-x" indicates the corpus type, where "x"
	 *            varies. The following parameters indicate the name(s) of input
	 *            file(s) and the output filename.
	 */
	public static void main(String[] args) {
		CorpusFormatter cf = new CorpusFormatter();
		switch (args[0].charAt(1)) {
		case 'o': // Omphalos
			cf.ReadOmphalosCorpus(args[1]);
			break;
		case 't': // Text
			cf.ReadTextCorpus(args[1]);
			break;
		}

		try {
			// store terms
			FileWriter fw = new FileWriter(args[args.length - 1]);
			fw.write(cf.terms.size() + "\n");
			for (String s : cf.terms) {
				fw.write(s + "\n");
			}
			// fw.close();

			// store sentences
			// fw = new FileWriter(args[args.length - 1]);
			fw.write("\n" + cf.sentences.size() + "\n");
			for (ArrayList<Integer> al : cf.sentences) {
				for (Integer i : al) {
					fw.write(i + " ");
				}
				fw.write("\n");
			}
			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	protected ArrayList<String> terms = new ArrayList<String>();
	protected ArrayList<ArrayList<Integer>> sentences = new ArrayList<ArrayList<Integer>>();

	protected void ReadOmphalosCorpus(String filename) {
		try {
			Scanner s = new Scanner(
					new BufferedReader(new FileReader(filename)));
			int sentNum = s.nextInt();
			int termNum = s.nextInt();
			HashMap<String, Integer> termMap = new HashMap<String, Integer>(
					(int) (termNum / 0.75 + 1));

			for (int i = 0; i < sentNum; i++) {
				if (s.nextInt() == 1) {
					int sentlen = s.nextInt();
					ArrayList<Integer> sent = new ArrayList<Integer>(sentlen);
					for (int j = 0; j < sentlen; j++) {
						String w = s.next();
						Integer wi = termMap.get(w);
						if (wi == null) {
							wi = terms.size();
							terms.add(w);
							termMap.put(w, wi);
						}
						sent.add(wi);
					}
					sentences.add(sent);
				}
			}

			s.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	protected void ReadTextCorpus(String filename) {
		try {
			Scanner s = new Scanner(
					new BufferedReader(new FileReader(filename)));
			HashMap<String, Integer> termMap = new HashMap<String, Integer>();
			while (s.hasNextLine()) {
				Scanner ss = new Scanner(s.nextLine());
				ArrayList<Integer> sent = new ArrayList<Integer>();
				while (ss.hasNext()) {
					String w = ss.next();
					Integer wi = termMap.get(w);
					if (wi == null) {
						wi = terms.size();
						terms.add(w);
						termMap.put(w, wi);
					}
					sent.add(wi);
				}
				sentences.add(sent);
			}
			s.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

}
