package gi.util;

import gi.AndNode;
import gi.Grammar;
import gi.OrNode;
import gi.Symbol;
import gi.Term;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Scanner;
import java.util.regex.MatchResult;

public class GrammarFormatter {

	/**
	 * @param args
	 *            The first parameter "-x" indicates the input grammar format,
	 *            where "x" varies. The following parameters indicate the
	 *            name(s) of input file(s) and the output filename.
	 */
	public static void main(String[] args) {
		switch (args[0].charAt(1)) {
		case 'l': // Lisp-like
			ReadLispLikeGrammar(args[1]);
			break;
		}

		if (g != null)
			g.saveToFile(args[args.length - 1]);
	}

	protected static Grammar g;

	protected static void ReadLispLikeGrammar(String filename) {
		HashMap<String, OrNode> orMap = new HashMap<String, OrNode>();
		HashMap<String, HashMap<String, AndNode>> andMap = new HashMap<String, HashMap<String, AndNode>>();
		HashMap<String, Term> termMap = new HashMap<String, Term>();
		try {
			Scanner s = new Scanner(
					new BufferedReader(new FileReader(filename)));
			s.skip("\\s*\\(");

			OrNode startSymbol = null;
			while (s.hasNext()) {
				s.skip("\\s*([\\(\\)])");
				String paren = s.match().group(1);
				if (paren.equals(")"))
					break;
				String nt = s.next();
				OrNode or = orMap.get(nt);
				if (or == null) {
					or = new OrNode(new ArrayList<Symbol>(),
							new ArrayList<double[]>());
					orMap.put(nt, or);
				}

				s.skip("\\s*->");

				while (true) {
					s.skip("\\s*([\\(\\)])");
					paren = s.match().group(1);
					if (paren.equals(")"))
						break;
					s.findWithinHorizon(
							"([\\d\\.]+)\\s*([^\\s\\(\\)]+)\\s*([^\\s\\(\\)]+)\\s*\\)", 0);
					MatchResult m = s.match();
					double weight = Double.valueOf(m.group(1));
					String left = m.group(2);
					String right = m.group(3);
					Symbol sym;
					if (left.equals(".")) {
						// terminal
						sym = termMap.get(right);
						if (sym == null) {
							Term t = new Term(right);
							termMap.put(right, t);
							sym = t;
						}
					} else {
						AndNode and = null;
						HashMap<String, AndNode> tmp = andMap.get(left);
						if (tmp != null)
							and = tmp.get(right);

						if (and == null) {
							OrNode lor = orMap.get(left);
							if (lor == null) {
								lor = new OrNode(new ArrayList<Symbol>(),
										new ArrayList<double[]>());
								orMap.put(left, lor);
							}
							OrNode ror = orMap.get(right);
							if (ror == null) {
								ror = new OrNode(new ArrayList<Symbol>(),
										new ArrayList<double[]>());
								orMap.put(right, ror);
							}
							and = new AndNode(lor, ror, 0);

							if (tmp == null) {
								tmp = new HashMap<String, AndNode>();
								andMap.put(left, tmp);
							}
							tmp.put(right, and);
						}

						sym = and;
					}
					or.children.add(sym);
					double[] tmp = { weight };
					or.weights.add(tmp);
					sym.parents.add(or);
				}

				if (startSymbol == null)
					startSymbol = or;
			}

			s.close();

			g = new Grammar(termMap.values().toArray(new Term[0]));
			for (HashMap<String, AndNode> m : andMap.values()) {
				for (AndNode and : m.values()) {
					g.andNodes.add(and);
				}
			}
			for (OrNode or : orMap.values()) {
				g.orNodes.add(or);
			}
			g.startSymbol = startSymbol;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

}
