package gi.util;

import gi.AndNode;
import gi.Corpus;
import gi.Grammar;
import gi.OrNode;
import gi.Sentence;
import gi.Symbol;
import gi.Term;
import gi.Sentence.Element;

import java.util.ArrayList;
import java.util.Arrays;

/**
 * Generate sentence(s) from a CNF-PCFG.
 * 
 * @author Kewei Tu
 * 
 */
public class SentenceGenerator {
	/**
	 * Generate a corpus
	 * 
	 * @param args
	 *            0: grammar file name; 1: number of sentences; 2: to-be-saved
	 *            corpus file name; 3: whether to show the generated sentences
	 *            on the screen.
	 */
	public static void main(String[] args) {
		SentenceGenerator sg = new SentenceGenerator();

		if (args.length == 4)
			sg.SHOW_SENT = Boolean.parseBoolean(args[3]);

		sg.grammar = new Grammar(args[0]);
		int size = Integer.valueOf(args[1]);
		ArrayList<Sentence> sentences = new ArrayList<Sentence>(size);
		for (int i = 0; i < size; i++) {
			sentences.add(sg.generate());
		}
		Corpus c = new Corpus(sentences.toArray(new Sentence[0]),
				sg.grammar.terms);
		c.saveToFile(args[2]);
	}

	protected Grammar grammar = null;

	public boolean SHOW_SENT = false;

	public void setGrammar(Grammar grammar) {
		this.grammar = grammar;
	}

	public Sentence generate() {
		if (grammar.startSymbol.children.size() == 0)
			return null;
		Sentence s = new Sentence(generatePart(grammar.startSymbol, 0)[0]);
		if (SHOW_SENT) {
			for (Element elem = s.firstElem; elem != null; elem = elem.next)
				System.out.print(elem.sym + " ");
			System.out.print("\n");
		}
		return s;
	}

	public static final int depthLimit = 50;

	/**
	 * Generate a list of Term from an OrNode.
	 * 
	 * @param or
	 * @return an array of two Element: the first and last Element of the
	 *         generated part.
	 */
	protected Element[] generatePart(OrNode or, int depth) {
		if (depth == depthLimit)
			System.out.println("oops, recursion too deep!");

		boolean tooDeep = depth > depthLimit;
		if (tooDeep) {
			ArrayList<Symbol> children = new ArrayList<Symbol>();
			ArrayList<double[]> weights = new ArrayList<double[]>();
			for (int i = 0; i < or.children.size(); i++) {
				Symbol s = or.children.get(i);
				if (s instanceof Term) {
					children.add(s);
					weights.add(or.weights.get(i));
				}
			}

			if (!children.isEmpty()) {
				double[] weightSums = new double[children.size()];
				double tmp = 0;
				for (int i = 0; i < weightSums.length; i++) {
					tmp = weightSums[i] = tmp + weights.get(i)[0];
				}
				double rand = Math.random() * tmp;
				int ind = Arrays.binarySearch(weightSums, rand);
				if (ind < 0)
					ind = -ind - 1;
				Symbol sym = children.get(ind);
				assert sym instanceof Term;
				Element e = new Element(sym);
				Element[] re = new Element[2];
				re[0] = re[1] = e;
				return re;
			}
		}

		double[] weightSums = new double[or.children.size()];
		double tmp = 0;
		for (int i = 0; i < weightSums.length; i++) {
			tmp = weightSums[i] = tmp + or.weights.get(i)[0];
		}
		double rand = Math.random() * tmp;
		int ind = Arrays.binarySearch(weightSums, rand);
		if (ind < 0)
			ind = -ind - 1;
		Symbol sym = or.children.get(ind);
		if (sym instanceof Term) {
			Element e = new Element(sym);
			Element[] re = new Element[2];
			re[0] = re[1] = e;
			return re;
		} else {
			AndNode and = (AndNode) sym;
			Element[] left = generatePart(and.left, depth + 1);
			Element[] right = generatePart(and.right, depth + 1);
			left[1].next = right[0];
			right[0].prev = left[1];
			left[1] = right[1];
			return left;
		}
	}
}
